"use strict";
// TODO: optional init function support
// TODO: class-validators support
Object.defineProperty(exports, "__esModule", { value: true });
const proxy_1 = require("./proxy");
const settings_1 = require("./settings");
const util_1 = require("./util");
const decorator_1 = require("./decorator");
function Mixin(...constructors) {
    const prototypes = constructors.map(constructor => constructor.prototype);
    function MixedClass(...args) {
        for (const constructor of constructors)
            util_1.copyProps(this, new constructor(...args));
    }
    MixedClass.prototype = settings_1.settings.prototypeStrategy === 'copy'
        ? util_1.hardMixProtos(prototypes, MixedClass)
        : util_1.softMixProtos(prototypes, MixedClass);
    Object.setPrototypeOf(MixedClass, settings_1.settings.staticsStrategy === 'copy'
        ? util_1.hardMixProtos(constructors, null, ['prototype'])
        : proxy_1.proxyMix(constructors, Function.prototype));
    let DecoratedMixedClass = MixedClass;
    for (let constructor of constructors) {
        const classDecorators = decorator_1.decorators.get(constructor);
        if (classDecorators) {
            if (classDecorators.class)
                for (let decorator of classDecorators.class)
                    DecoratedMixedClass = decorator(DecoratedMixedClass);
            if (classDecorators.static)
                applyPropAndMethodDecorators(classDecorators.static, DecoratedMixedClass);
            if (classDecorators.instance)
                applyPropAndMethodDecorators(classDecorators.instance, DecoratedMixedClass.prototype);
        }
    }
    return DecoratedMixedClass;
}
exports.Mixin = Mixin;
const applyPropAndMethodDecorators = (propAndMethodDecorators, target) => {
    const propDecorators = propAndMethodDecorators.property;
    const methodDecorators = propAndMethodDecorators.method;
    if (propDecorators)
        for (let key in propDecorators)
            for (let decorator of propDecorators[key])
                decorator(target, key);
    if (methodDecorators)
        for (let key in methodDecorators)
            for (let decorator of methodDecorators[key])
                decorator(target, key, Object.getOwnPropertyDescriptor(target, key));
};
/**
 * A decorator version of the `Mixin` function.  You'll want to use this instead of `Mixin` for mixing generic classes.
 */
const mix = (...ingredients) => 
// @ts-ignore
decoratedClass => Mixin(...(ingredients.concat([decoratedClass])));
exports.mix = mix;
